import json
import requests


def api_method(method):
    """Decorator for using method signatures for generating url endpoint."""
    def wrapper(self, *args, **kwargs):
        method(self, *args, **kwargs)
        api = self.api
        vehicle = self.vehicle
        vehicle_id = '&'.join(['{}={}'.format(key, value) for key, value in vehicle.items() if value])
        if not vehicle_id:
            endpoint = api
        else:
            endpoint = '{}?{}'.format(api, vehicle_id)
        response = self.get(endpoint)

        return response

    return wrapper


class BaseAPI():
    """Base wraper for individual CarMD requests."""
    base_url = 'http://api.carmd.com/v3.0'
    api_header = {
        "content-type": "application/json",
        "authorization": "your_auth_key_here",
        "partner-token": "your_partner_token_here"
    }

    def __init__(self, auth, partner, **kwargs):
        """Initialize with appropriate authorization and partner-token for API header"""
        if not auth:
            raise KeyError("Please pass a valid authorization key: "
                           "e.g. 'Basic MmE0MzBkZjYtOTIxOS00ODhjLTllMjktNjQ2MDlhMmY1OWZx'")
        if not partner:
            raise KeyError("Please pass a valid partner-token key: "
                           "e.g. 'e18dc0f62dfb456398c83b893d3e81e2'")
        self.api_header["authorization"] = auth
        self.api_header["partner-token"] = partner
        self.kwargs = kwargs

    def get(self, endpoint):
        """Get request from specified url endpoint."""
        url = '{}/{}'.format(self.base_url, endpoint)
        response = requests.get(url, headers=self.api_header)
        if response.status_code != 200:
            raise ConnectionError("Bad request: {}".format(response.status_code))
        response_json = json.loads(response.content)

        return response_json


class FieldsVin(BaseAPI):
    """API for available fields under VIN:
    http://api.carmd.com/v3.0/fields"""
    api = 'fields'

    @api_method
    def __call__(self, mileage=None, dtc=None):
        self.vehicle = self.kwargs['vehicle']
        self.vehicle['mileage'] = mileage
        self.vehicle['dtc'] = dtc


class FieldsMake(BaseAPI):
    """API for available fields under make:
    http://api.carmd.com/v3.0/fields"""
    api = 'fields'

    @api_method
    def __call__(self, mileage=None):
        self.vehicle = self.kwargs['vehicle']
        self.vehicle['mileage'] = mileage


class Decode(BaseAPI):
    """API for VIN decode:
    http://api.carmd.com/v3.0/decode"""
    api = 'decode'

    @api_method
    def __call__(self):
        self.vehicle = self.kwargs['vehicle']


class DecodeEnhanced(BaseAPI):
    """
    Coming soon - to be an enhanced version
    of VIN decode.
    """
    # api = 'decode_enh'
    pass


class OBDPortLocation(BaseAPI):
    """API for finding OBD2 port location:
    http://api.carmd.com/v3.0/port"""
    api = 'port'

    @api_method
    def __call__(self):
        self.vehicle = self.kwargs['vehicle']


class Maintenance(BaseAPI):
    """API for maintenance:
    http://api.carmd.com/v3.0/maint"""
    api = 'maint'

    @api_method
    def __call__(self, mileage):
        self.vehicle = self.kwargs['vehicle']
        self.vehicle['mileage'] = mileage


class MaintenanceList(BaseAPI):
    """API for full maintenance list:
    http://api.carmd.com/v3.0/maintlist"""
    api = 'maintlist'

    @api_method
    def __call__(self):
        self.vehicle = self.kwargs['vehicle']


class Repairs(BaseAPI):
    """API for repair information:
    http://api.carmd.com/v3.0/repair"""
    api = 'repair'

    @api_method
    def __call__(self, mileage, dtc):
        self.vehicle = self.kwargs['vehicle']
        self.vehicle['mileage'] = mileage
        self.vehicle['dtc'] = dtc


class Diagnostics(BaseAPI):
    """API for diagnostic information:
    http://api.carmd.com/v3.0/diag"""
    api = 'diag'

    @api_method
    def __call__(self, mileage, dtc):
        self.vehicle = self.kwargs['vehicle']
        self.vehicle['mileage'] = mileage
        self.vehicle['dtc'] = dtc


class UpcomingRepairs(BaseAPI):
    """API for upcoming repairs:
    http://api.carmd.com/v3.0/upcoming"""
    api = 'upcoming'

    @api_method
    def __call__(self, mileage):
        self.vehicle = self.kwargs['vehicle']
        self.vehicle['mileage'] = mileage


class TSBVin(BaseAPI):
    """API for technical service bulletin using VIN:
    http://api.carmd.com/v3.0/tsb"""
    api = 'tsb'

    @api_method
    def __call__(self):
        self.vehicle = self.kwargs['vehicle']


class TSBMake(BaseAPI):
    """API for technical service bulletin using make:
    http://api.carmd.com/v3.0/tsb"""
    api = 'tsb'

    @api_method
    def __call__(self, engine):
        self.vehicle = self.kwargs['vehicle']
        self.vehicle['engine'] = engine


class SafetyRecalls(BaseAPI):
    """API for safety recalls:
    http://api.carmd.com/v3.0/recall"""
    api = 'recall'

    @api_method
    def __call__(self):
        self.vehicle = self.kwargs['vehicle']


class VehicleWarranty(BaseAPI):
    """API for warranty status:
    http://api.carmd.com/v3.0/warranty"""
    api = 'warranty'

    @api_method
    def __call__(self):
        self.vehicle = self.kwargs['vehicle']


class VehicleImage(BaseAPI):
    """API for getting vehicle image:
    http://api.carmd.com/v3.0/image"""
    api = 'image'

    @api_method
    def __call__(self):
        self.vehicle = self.kwargs['vehicle']


class YMME(BaseAPI):
    """API for getting year, model, make, and engine:
    http://api.carmd.com/v3.0/year
    http://api.carmd.com/v3.0/make
    http://api.carmd.com/v3.0/model
    http://api.carmd.com/v3.0/engine"""
    @api_method
    def year(self):
        self.api = 'year'
        self.vehicle = {}

    @api_method
    def make(self, year):
        self.api = 'make'
        self.vehicle = {
            'year': year
        }

    @api_method
    def model(self, year, make):
        self.api = 'model'
        self.vehicle = {
            'year': year,
            'make': make.upper()
        }

    @api_method
    def engine(self, year, make, model):
        self.api = 'engine'
        self.vehicle = {
            'year': year,
            'make': make.upper(),
            'model': model.upper()
        }


class Credits(BaseAPI):
    """API for account credits:
    http://api.carmd.com/v3.0/credits"""
    api = 'credits'

    @api_method
    def __call__(self):
        self.vehicle = {}


class CarMD():
    """General class for the CarMD API"""
    def __init__(self, auth_key=None, partner_key=None):
        self.args = (auth_key, partner_key)
        self.ymme = YMME(*self.args)
        self.acct_credits = Credits(*self.args)
        self.__vehicle = {}

    def __str__(self):
        return "CarMD Object - Vehicle ID: {}".format(self.__vehicle)

    def vin(self, vin):
        """Assign a VIN to the CarMD() instance"""
        self.my_vehicle = {
            'vin': vin}
        self.__vehicle = self.my_vehicle
        self.instantiate(vehicle=self.my_vehicle)
        return self

    def make(self, year, make, model):
        """Assign a make to the CarMD() instance"""
        self.my_vehicle = {
            'year': int(year),
            'make': make.upper(),
            'model': model.upper()
        }
        self.__vehicle = self.my_vehicle
        self.instantiate(vehicle=self.my_vehicle)
        return self

    def instantiate(self, **kwargs):
        """Instantiate appropriate classes by VIN or make"""
        if 'vin' in kwargs['vehicle']:
            self.fields = FieldsVin(*self.args, **kwargs)
            self.tech_service = TSBVin(*self.args, **kwargs)
            self.decode = Decode(*self.args, **kwargs)
            self.decode_enhanced = DecodeEnhanced(*self.args, **kwargs)
            self.repairs = Repairs(*self.args, **kwargs)
            self.diagnostics = Diagnostics(*self.args, **kwargs)
            self.vehicle_image = VehicleImage(*self.args, **kwargs)
        else:
            self.fields = FieldsMake(*self.args, **kwargs)
            self.tech_service = TSBMake(*self.args, **kwargs)
        self.obd2 = OBDPortLocation(*self.args, **kwargs)
        self.maintenance = Maintenance(*self.args, **kwargs)
        self.maintenance_list = MaintenanceList(*self.args, **kwargs)
        self.future_repairs = UpcomingRepairs(*self.args, **kwargs)
        self.recalls = SafetyRecalls(*self.args, **kwargs)
        self.warranty = VehicleWarranty(*self.args, **kwargs)



"""
Unit testing for carmd module in ../carmd
"""

import os
from pathlib import Path
import sys
import unittest
os.chdir(Path(__file__).parent)
os.chdir('../carmd')
sys.path.append(os.getcwd())  # required for relative file fetching - run in 'test' directory
from carmd import CarMD


class TestCarMD(unittest.TestCase):
    AUTHORIZATION = 'Basic MmE0MzBkZjYtOTIxOS00ODhjLTllMjktNjQ2MDlhMmY1OWZi'
    PARTNER_TOKEN = 'e18dc0f62dfb456398c83b893d3e81e1'
    vin = "1GNALDEK9FZ108495"
    year = 2015
    make = "chevrolet"
    model = "equinox"
    engine = "L4,2.4L;DOHC;16V;DI;FFV"
    mileage = 125000
    dtc = "p0420"

    def setUp(self):
        # comment out if you want to view instance influence on methods
        self.carmd = CarMD(self.AUTHORIZATION, self.PARTNER_TOKEN)

    def test_fields_vin(self):
        vehicle_vin = self.carmd.vin(self.vin)
        self.assertIsInstance(vehicle_vin.fields(self.mileage, self.dtc), dict)

    def test_fields_make(self):
        vehicle_make = self.carmd.make(self.year, self.make, self.model)
        self.assertIsInstance(vehicle_make.fields(), dict)
        self.assertIsInstance(vehicle_make.fields(self.mileage), dict)
        with self.assertRaises(TypeError):
            vehicle_make.fields(self.mileage, dtc=self.dtc), dict

    def test_decode(self):
        vehicle_vin = self.carmd.vin(self.vin)
        self.assertIsInstance(vehicle_vin.decode(), dict)

    def test_obd2_vin(self):
        vehicle_vin = self.carmd.vin(self.vin)
        self.assertIsInstance(vehicle_vin.obd2(), dict)

    def test_obd2_make(self):
        vehicle_make = self.carmd.make(self.year, self.make, self.model)
        self.assertIsInstance(vehicle_make.obd2(), dict)

    def test_maintenance_vin(self):
        vehicle_vin = self.carmd.vin(self.vin)
        self.assertIsInstance(vehicle_vin.maintenance(self.mileage), dict)
        with self.assertRaises(TypeError):
            vehicle_vin.maintenance()

    def test_maintenance_make(self):
        vehicle_make = self.carmd.make(self.year, self.make, self.model)
        self.assertIsInstance(vehicle_make.maintenance(self.mileage), dict)
        with self.assertRaises(TypeError):
            vehicle_make.maintenance()

    def test_maintenance_list_vin(self):
        vehicle_vin = self.carmd.vin(self.vin)
        self.assertIsInstance(vehicle_vin.maintenance_list(), dict)

    def test_maintenance_list_make(self):
        vehicle_make = self.carmd.make(self.year, self.make, self.model)
        self.assertIsInstance(vehicle_make.maintenance_list(), dict)

    def test_repair(self):
        vehicle_vin = self.carmd.vin(self.vin)
        self.assertIsInstance(vehicle_vin.repairs(mileage=self.mileage, dtc=self.dtc), dict)
        with self.assertRaises(TypeError):
            vehicle_vin.repairs()
        with self.assertRaises(TypeError):
            vehicle_vin.repairs(mileage=self.mileage)
        with self.assertRaises(TypeError):
            vehicle_vin.repairs(dtc=self.dtc)

    def test_diagnostics(self):
        vehicle_vin = self.carmd.vin(self.vin)
        self.assertIsInstance(vehicle_vin.diagnostics(mileage=self.mileage, dtc=self.dtc), dict)
        with self.assertRaises(TypeError):
            vehicle_vin.diagnostics()
        with self.assertRaises(TypeError):
            vehicle_vin.diagnostics(mileage=self.mileage)
        with self.assertRaises(TypeError):
            vehicle_vin.diagnostics(dtc=self.dtc)

    def test_future_repairs_vin(self):
        vehicle_vin = self.carmd.vin(self.vin)
        self.assertIsInstance(vehicle_vin.future_repairs(mileage=self.mileage), dict)
        with self.assertRaises(TypeError):
            vehicle_vin.future_repairs()

    def test_future_repairs_make(self):
        vehicle_make = self.carmd.make(self.year, self.make, self.model)
        self.assertIsInstance(vehicle_make.future_repairs(mileage=self.mileage), dict)
        with self.assertRaises(TypeError):
            vehicle_make.future_repairs()

    def test_tech_service_vin(self):
        vehicle_vin = self.carmd.vin(self.vin)
        self.assertIsInstance(vehicle_vin.tech_service(), dict)

    def test_tech_service_make(self):
        vehicle_make = self.carmd.make(self.year, self.make, self.model)
        self.assertIsInstance(vehicle_make.tech_service(engine=self.engine), dict)
        with self.assertRaises(TypeError):
            vehicle_make.tech_service()

    def test_recalls_vin(self):
        vehicle_vin = self.carmd.vin(self.vin)
        self.assertIsInstance(vehicle_vin.recalls(), dict)

    def test_recalls_make(self):
        vehicle_make = self.carmd.make(self.year, self.make, self.model)
        self.assertIsInstance(vehicle_make.recalls(), dict)

    def test_warranty_vin(self):
        vehicle_vin = self.carmd.vin(self.vin)
        self.assertIsInstance(vehicle_vin.warranty(), dict)

    def test_warranty_make(self):
        vehicle_make = self.carmd.make(self.year, self.make, self.model)
        self.assertIsInstance(vehicle_make.warranty(), dict)

    def test_vehicle_image(self):
        vehicle_vin = self.carmd.vin(self.vin)
        self.assertIsInstance(vehicle_vin.vehicle_image(), dict)

    def test_year(self):
        self.assertIsInstance(self.carmd.ymme.year(), dict)
        vehicle_vin = self.carmd.vin(self.vin)
        self.assertIsInstance(vehicle_vin.ymme.year(), dict)
        vehicle_make = self.carmd.make(self.year, self.make, self.model)
        self.assertIsInstance(vehicle_make.ymme.year(), dict)

    def test_make(self):
        self.assertIsInstance(self.carmd.ymme.make(self.year), dict)
        vehicle_vin = self.carmd.vin(self.vin)
        self.assertIsInstance(vehicle_vin.ymme.make(self.year), dict)
        vehicle_make = self.carmd.make(self.year, self.make, self.model)
        self.assertIsInstance(vehicle_make.ymme.make(self.year), dict)

    def test_model(self):
        self.assertIsInstance(self.carmd.ymme.model(self.year, self.make), dict)
        vehicle_vin = self.carmd.vin(self.vin)
        self.assertIsInstance(vehicle_vin.ymme.model(self.year, self.make), dict)
        vehicle_make = self.carmd.make(self.year, self.make, self.model)
        self.assertIsInstance(vehicle_make.ymme.model(self.year, self.make), dict)

    def test_engine(self):
        self.assertIsInstance(self.carmd.ymme.engine(self.year, self.make, self.model), dict)
        vehicle_vin = self.carmd.vin(self.vin)
        self.assertIsInstance(vehicle_vin.ymme.engine(self.year, self.make, self.model), dict)
        vehicle_make = self.carmd.make(self.year, self.make, self.model)
        self.assertIsInstance(vehicle_make.ymme.engine(self.year, self.make, self.model), dict)

    def test_credits(self):
        self.assertIsInstance(self.carmd.acct_credits(), dict)
        vehicle_vin = self.carmd.vin(self.vin)
        self.assertIsInstance(vehicle_vin.acct_credits(), dict)
        vehicle_make = self.carmd.make(self.year, self.make, self.model)
        self.assertIsInstance(vehicle_make.acct_credits(), dict)

    def test_non_instance(self):
        with self.assertRaises(KeyError):
            CarMD()

    def test_partial_instance(self):
        with self.assertRaises(KeyError):
            CarMD('Basic MmE0MzBkZjYtOTIxOS00ODhjLTllMjktNjQ2MDlhMmY1OWZi')

    def test_connection_error(self):
        with self.assertRaises(ConnectionError):
            self.carmd.ymme.make(self.model)  # accepts self.year


if __name__ == '__main__':
    unittest.main()


"""Setup for carmd package."""

import setuptools


with open('README.rst') as f:
    README = f.read()

setuptools.setup(
    author="Ira Horecka",
    author_email="ira89@icloud.com",
    name='python-carmd',
    license="MIT",
    description='A simple CarMD API wrapper',
    version='v0.2.0',
    long_description=README,
    long_description_content_type="text/x-rst",
    url='https://github.com/irahorecka/python-carmd',
    packages=['carmd'],
    python_requires=">=3.5",
    install_requires=['requests'],
    classifiers=[
        # Trove classifiers
        # (https://pypi.python.org/pypi?%3Aaction=list_classifiers)
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Intended Audience :: Developers',
    ],
)
